#include <os.h>
#include "disp.h"
#include "WELL1024.h"
#include "oldsprites.h"

#define RTC_POINTER ((unsigned int*)0x90090000)
#define FLOORY 40
#define FLOORTHICKNESS BLOCKHEIGHTWIDTH

#define RUNNERX 40

#define BLOCKGROUPNUM 5

uint8_t blockgroups[BLOCKGROUPNUM][2] = {{0x00,0x00},{0x1C,0x00},{0x10,0x00},{0x00,0x08},{0x00,0x18}};//blockgroupformat: 8 blockmgl, 2 bits pro x-position -> jede blockgruppe braucht fr obere reihe ein byte und fr untere eins, oberes [][0] unteres [][1]

inline int drawrunnersprite(uint8_t* bufaddr, uint8_t* spraddr, int height, int width){
    int row, doublecol, retval=0;
    for(row=0;row<height;row++){
        for(doublecol=0;doublecol<width/2;doublecol++){
            if(!(*bufaddr) && (*spraddr!=0xFF))//if runner touches black pixel
                retval=1;
            if(*spraddr != 0xFF)
                *bufaddr = *spraddr;
            spraddr++;
            bufaddr++;
        }
        bufaddr += (SCREEN_WIDTH-width)/2;
    }
    return retval;
}

int drawrunner(uint8_t* scrbuf, signed char jumpheight){
    uint8_t* bufaddr;
	uint8_t* spraddr;
	static int animationnum = 0;
	if(jumpheight>0){
	    bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-RUNNERHEIGHT-jumpheight)*SCREEN_WIDTH/2 + RUNNERX/2;
	    spraddr = linerunner_jumping[0];
        return drawrunnersprite(bufaddr,spraddr,RUNNERHEIGHT,RUNNERWIDTH);
	}
	else if(jumpheight==0){
	    bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-RUNNERHEIGHT)*SCREEN_WIDTH/2 + RUNNERX/2;
	    spraddr = runner[animationnum/3];
        animationnum++;
        if(animationnum == RUNNING_RUNNER_FRAMES*3)
            animationnum=0;
        return drawrunnersprite(bufaddr,spraddr,RUNNERHEIGHT,RUNNERWIDTH);
	}
	else{
	    bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH)*SCREEN_WIDTH/2 + RUNNERX/2;
	    spraddr = linerunner_rolling[0];
	    return drawrunnersprite(bufaddr,spraddr,BLOCKHEIGHTWIDTH,BLOCKHEIGHTWIDTH);
	}
}

void drawgameover(uint8_t* scrbuf){
    uint8_t* bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-GRAVESTONEHEIGHT)*SCREEN_WIDTH/2 + RUNNERX/2;
	uint8_t* spraddr = dead;
	int row, doublecol;
	for(row=0;row<GRAVESTONEHEIGHT;row++){
	    for(doublecol=0;doublecol<GRAVESTONEWIDTH/2;doublecol++){
            if(*spraddr != 0xFF)
                *bufaddr = *spraddr;
            spraddr++;
            bufaddr++;
	    }
	    bufaddr += (SCREEN_WIDTH-GRAVESTONEWIDTH)/2;
	}
}

void updateallblocks(uint8_t* blockscrbuf, uint64_t currentblockgroups[2], char offset){
    int i;
    for(i=0;i<BLOCKHEIGHTWIDTH;i++)//move high blocks 2 pixels to the left
        memmove(blockscrbuf+(SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE+i)*SCREEN_WIDTH/2, blockscrbuf+(SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE+i)*SCREEN_WIDTH/2 + 1, SCREEN_WIDTH/2-1);
    for(i=0;i<BLOCKHEIGHTWIDTH;i++)//move low blocks 2 pixels to the left
        memmove(blockscrbuf+(SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2, blockscrbuf+(SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2 + 1, SCREEN_WIDTH/2-1);
    if(offset && (currentblockgroups[0] & 0x0000000080000000) && (offset<=BLOCKHEIGHTWIDTH)){//high blocks check
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            *(uint8_t*)(blockscrbuf + ((SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE+i) * (SCREEN_WIDTH/2)) + (SCREEN_WIDTH/2) - 1) = 0x00;//set rightmost pixel
    }
    else{
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            *(uint8_t*)(blockscrbuf + ((SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE+i) * (SCREEN_WIDTH/2)) + (SCREEN_WIDTH/2) - 1) = 0xFF;//delete rightmost pixel
    }
    if(offset && (currentblockgroups[1] & 0x0000000080000000) && (offset<=BLOCKHEIGHTWIDTH)){//low blocks check
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            *(uint8_t*)(blockscrbuf + ((SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i) * (SCREEN_WIDTH/2)) + (SCREEN_WIDTH/2) - 1) = 0x00;//set rightmost pixel
    }
    else{
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            *(uint8_t*)(blockscrbuf + ((SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i) * (SCREEN_WIDTH/2)) + (SCREEN_WIDTH/2) - 1) = 0xFF;//delete rightmost pixel
    }
}

void reload(uint64_t currentblockgroups[2]){
    short group;
    static char high = 0, low = 0, empty = 0;
    int i;
    for(i=0;i<500;i++){
        group = WELLRNG1024a()%BLOCKGROUPNUM;
        if(group > 0 && group < 3 && high < 2){//high group and less than 2 high groups before
            high++;
            low = 0;
            empty = 0;
            break;
        }
        if(group == 0 && empty < 5){
            empty++;
            break;
        }
        if(group > 2 && low < 4){
            low++;
            empty = 0;
            high = 0;
            break;
        }
    }
    currentblockgroups[0] += blockgroups[group][0];
    currentblockgroups[1] += blockgroups[group][1];
}

int main(){
    clrscr();
    lcd_ingray();
    InitWELLRNG1024a((unsigned int*)((*(volatile unsigned*)0xC0000010) + 120 * 160 + 32));
    uint8_t* scrbuf = (uint8_t*) malloc(SCREEN_BYTES_SIZE);
    uint8_t* blockscrbuf = (uint8_t*) malloc(SCREEN_BYTES_SIZE);
    signed char jumpheight=0, canjump = 1, rolling = 0, canroll=0;
    uint64_t currentblockgroups[2];
    char gameover=0;
    char offset=0;
    char loadgroups = 0;
    int score=0;
    int speed=16;//milliseconds to sleep -> the lower, the faster
    int i;
    char scorestring[6];
    currentblockgroups[0] = 0x0000000001000000;//obere reihe
    currentblockgroups[1] = 0x0000001C00010000;//untere reihe
    reload(currentblockgroups);//load first group
    memset(blockscrbuf, 0xFF, SCREEN_BYTES_SIZE);
    memset(scrbuf, 0xFF, SCREEN_BYTES_SIZE);
    memset(blockscrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS)*SCREEN_WIDTH/2, 0x00, FLOORTHICKNESS*SCREEN_WIDTH/2);//draw floor
    int fps=0;
    int lasttime = 0;
    while(!isKeyPressed(KEY_NSPIRE_ESC)){
        sleep(speed);
        memset(scrbuf, 0xFF, SCREEN_BYTES_SIZE);
        sprintf(scorestring, "%i", score);
        dispString(scrbuf, 0, 0, scorestring);
        updateallblocks(blockscrbuf, currentblockgroups, offset);
        memcpy(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE)*SCREEN_WIDTH/2, blockscrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-2*BLOCKHEIGHTWIDTH-BLOCKYDISTANCE)*SCREEN_WIDTH/2, (FLOORTHICKNESS+2*BLOCKHEIGHTWIDTH+BLOCKYDISTANCE)*SCREEN_WIDTH/2);
        gameover = drawrunner(scrbuf, jumpheight);
        //if(is_cx)
            memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
        if(*(int*)0x90090000 == lasttime)
            fps++;
        else{
            printf("%i\n",fps);
            fps=0;
            lasttime = *(int*)0x90090000;
        }
        //else if(!(offset%8))//reduce fps on monochrome nspires
        //    memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
        offset+=2;
        if(offset == BLOCKHEIGHTWIDTH+BLOCKXDISTANCE){
            offset=0;
            currentblockgroups[0]<<=1;
            currentblockgroups[1]<<=1;
            loadgroups++;
            score++;
            if(!(score%100)){
                if(score == 100)
                    speed--;
                else if(score == 200)
                    speed--;
                else if(score == 300)
                    speed--;
                else if(score == 500)
                    speed--;
                else if(score == 700)
                    speed--;
                else if(score == 1000)
                    speed--;
                else if(score == 1400)
                    speed--;
                else if(score == 1900)
                    speed--;
                else if(score == 2500)
                    speed--;
                else if(score == 3000)
                    speed--;
                else if(score == 5000)
                    speed--;
                else if(score == 10000)
                    speed--;
                else if(score == 20000)
                    speed--;
                else if(score == 30000)
                    speed--;
                else if(score == 40000)
                    speed--;
                else if(score == 50000)
                    speed--;
                else if(score == 70000)
                    speed--;
            }
        }
        if(score == 99999){
            dispString(scrbuf, 110, 100, "You win!");//to prevent overflow of string, todo: change to bigger number :)
            memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
            wait_no_key_pressed();
            wait_key_pressed();
            break;
        }
        if(loadgroups == 8){//if only 56 bits/blocks left, reload to 64
            reload(currentblockgroups);
            loadgroups = 0;
        }
        if(isKeyPressed(KEY_NSPIRE_LEFT) || isKeyPressed(KEY_NSPIRE_UP)){
            if(jumpheight <= 0)
                canjump = 1;
            if(jumpheight<0)
                jumpheight=0;
            if((jumpheight<4*BLOCKHEIGHTWIDTH) && (canjump == 1))
                jumpheight+=BLOCKHEIGHTWIDTH;
            else//if cannot jump or at highest height
                canjump = 0;
        }
        else if(jumpheight>0){
            canjump = 0;
        }
        if((canroll == 1) && (isKeyPressed(KEY_NSPIRE_RIGHT) || isKeyPressed(KEY_NSPIRE_DOWN))){
            jumpheight = -1;
            canroll = 0;
            rolling = 8*BLOCKHEIGHTWIDTH;
        }
        if(jumpheight == -1){
            rolling -= 2;
            if(rolling <= 0){
                jumpheight=0;
                canroll = 0;
            }
        }
        else{
            canroll = 1;
        }
        if(jumpheight>0){
            if(jumpheight>1)
                jumpheight -= 1;
            else
                jumpheight = 0;
        }
        if(gameover){
            memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
            for(i=0;i<10;i++){
                sleep(20);
                fade(scrbuf, 1);
                drawgameover(scrbuf);
                memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
            }
            wait_no_key_pressed();
            wait_key_pressed();
            break;
        }
    }
    free(scrbuf);
    free(blockscrbuf);
    return 0;
}
